@@ -138,6 +138,8 @@ static void parse_preamble(struct parser *p) {
 			}
 			fprintf(p->output, "\n");
 			break;
+		} else if (section == -1) {
+			parser_fatal(p, "Name characters must be A-Z, a-z, 0-9, `-`, `_`, or `.`");
 		}
 	}
 	str_free(name);
@@ -195,7 +197,7 @@ static void parse_linebreak(struct parser *p) {
 }
 
 static void parse_text(struct parser *p) {
-	uint32_t ch, last = ' ';
+	uint32_t ch, next, last = ' ';
 	int i = 0;
 	while ((ch = parser_getch(p)) != UTF8_INVALID) {
 		switch (ch) {
@@ -213,11 +215,16 @@ static void parse_text(struct parser *p) {
 			parse_format(p, FORMAT_BOLD);
 			break;
 		case '_':
-			if (!isalnum(last) || (p->flags & FORMAT_UNDERLINE)) {
+			next = parser_getch(p);
+			if (!isalnum(last) || ((p->flags & FORMAT_UNDERLINE) && !isalnum(next))) {
 				parse_format(p, FORMAT_UNDERLINE);
 			} else {
 				utf8_fputch(p->output, ch);
 			}
+			if (next == UTF8_INVALID) {
+				return;
+			}
+			parser_pushch(p, next);
 			break;
 		case '+':
 			parse_linebreak(p);
@@ -279,19 +286,19 @@ static int parse_indent(struct parser *p, int *indent, bool write) {
 		++i;
 	}
 	parser_pushch(p, ch);
-	if (ch == '\n' && *indent != 0) {
-		// Don't change indent when we encounter empty lines
+	if ((ch == '\n' || ch == UTF8_INVALID) && *indent != 0) {
+		// Don't change indent when we encounter empty lines or EOF
 		return *indent;
 	}
 	if (write) {
-		if (i < *indent) {
+		if ((i - *indent) > 1) {
+			parser_fatal(p, "Indented by an amount greater than 1");
+		} else if (i < *indent) {
 			for (int j = *indent; i < j; --j) {
 				roff_macro(p, "RE", NULL);
 			}
 		} else if (i == *indent + 1) {
 			fprintf(p->output, ".RS 4\n");
-		} else if (i != *indent && ch == '\t') {
-			parser_fatal(p, "Indented by an amount greater than 1");
 		}
 	}
 	*indent = i;
